package prefuse.data;

import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import prefuse.data.column.Column;
import prefuse.data.column.ColumnFactory;
import prefuse.data.column.ColumnMetadata;
import prefuse.data.event.ColumnListener;
import prefuse.data.event.TableListener;
import prefuse.data.expression.Expression;
import prefuse.data.expression.Predicate;
import prefuse.data.expression.parser.ExpressionParser;
import prefuse.data.tuple.AbstractTupleSet;
import prefuse.data.tuple.TableTuple;
import prefuse.data.tuple.TupleManager;
import prefuse.data.util.FilterIteratorFactory;
import prefuse.data.util.Index;
import prefuse.data.util.RowManager;
import prefuse.data.util.TableIterator;
import prefuse.data.util.TreeIndex;
import prefuse.util.TypeLib;
import prefuse.util.collections.CopyOnWriteArrayList;
import prefuse.util.collections.IncompatibleComparatorException;
import prefuse.util.collections.IntIterator;

public class Table
  extends AbstractTupleSet
  implements ColumnListener
{
  protected CopyOnWriteArrayList m_listeners = new CopyOnWriteArrayList();
  protected ArrayList m_columns;
  protected ArrayList m_names;
  protected HashMap m_entries;
  protected RowManager m_rows;
  protected TupleManager m_tuples;
  protected int m_modCount = 0;
  protected int m_lastCol = -1;
  protected Schema m_schema;
  
  public Table()
  {
    this(0, 0);
  }
  
  public Table(int paramInt1, int paramInt2)
  {
    this(paramInt1, paramInt2, TableTuple.class);
  }
  
  protected Table(int paramInt1, int paramInt2, Class paramClass)
  {
    this.m_columns = new ArrayList(paramInt2);
    this.m_names = new ArrayList(paramInt2);
    this.m_rows = new RowManager(this);
    this.m_entries = new HashMap(paramInt2 + 5);
    this.m_tuples = new TupleManager(this, null, paramClass);
    if (paramInt1 > 0) {
      addRows(paramInt1);
    }
  }
  
  public int getColumnCount()
  {
    return this.m_columns.size();
  }
  
  public Class getColumnType(int paramInt)
  {
    return getColumn(paramInt).getColumnClass();
  }
  
  public Class getColumnType(String paramString)
  {
    Column localColumn = getColumn(paramString);
    return localColumn == null ? null : localColumn.getColumnClass();
  }
  
  public int getRowCount()
  {
    return this.m_rows.getRowCount();
  }
  
  public boolean isCellEditable(int paramInt1, int paramInt2)
  {
    if (!this.m_rows.isValidRow(paramInt1)) {
      return false;
    }
    return getColumn(paramInt2).isCellEditable(paramInt1);
  }
  
  public int getModificationCount()
  {
    return this.m_modCount;
  }
  
  public void setTupleManager(TupleManager paramTupleManager)
  {
    this.m_tuples.invalidateAll();
    this.m_tuples = paramTupleManager;
  }
  
  public Schema getSchema()
  {
    if (this.m_schema == null)
    {
      Schema localSchema = new Schema();
      for (int i = 0; i < getColumnCount(); i++) {
        localSchema.addColumn(getColumnName(i), getColumnType(i), getColumn(i).getDefaultValue());
      }
      localSchema.lockSchema();
      this.m_schema = localSchema;
    }
    return this.m_schema;
  }
  
  protected void invalidateSchema()
  {
    this.m_schema = null;
  }
  
  public int getColumnRow(int paramInt1, int paramInt2)
  {
    return this.m_rows.getColumnRow(paramInt1, paramInt2);
  }
  
  public int getTableRow(int paramInt1, int paramInt2)
  {
    return this.m_rows.getTableRow(paramInt1, paramInt2);
  }
  
  public int addRow()
  {
    int i = this.m_rows.addRow();
    updateRowCount();
    fireTableEvent(i, i, -1, 1);
    return i;
  }
  
  public void addRows(int paramInt)
  {
    for (int i = 0; i < paramInt; i++) {
      addRow();
    }
  }
  
  protected void updateRowCount()
  {
    int i = this.m_rows.getMaximumRow() + 1;
    this.m_tuples.setMaximumRow(i);
    Iterator localIterator = getColumns();
    while (localIterator.hasNext())
    {
      Column localColumn = (Column)localIterator.next();
      localColumn.setMaximumRow(i);
    }
  }
  
  public boolean removeRow(int paramInt)
  {
    if (this.m_rows.isValidRow(paramInt))
    {
      fireTableEvent(paramInt, paramInt, -1, -1);
      this.m_tuples.invalidate(paramInt);
      this.m_rows.releaseRow(paramInt);
      Iterator localIterator = getColumns();
      while (localIterator.hasNext())
      {
        Column localColumn = (Column)localIterator.next();
        localColumn.revertToDefault(paramInt);
      }
      return true;
    }
    return false;
  }
  
  public void clear()
  {
    IntIterator localIntIterator = rows(true);
    while (localIntIterator.hasNext()) {
      removeRow(localIntIterator.nextInt());
    }
  }
  
  public boolean isValidRow(int paramInt)
  {
    return this.m_rows.isValidRow(paramInt);
  }
  
  protected boolean hasColumn(String paramString)
  {
    return getColumnNumber(paramString) != -1;
  }
  
  public String getColumnName(int paramInt)
  {
    return (String)this.m_names.get(paramInt);
  }
  
  public int getColumnNumber(String paramString)
  {
    ColumnEntry localColumnEntry = (ColumnEntry)this.m_entries.get(paramString);
    return localColumnEntry == null ? -1 : localColumnEntry.colnum;
  }
  
  public int getColumnNumber(Column paramColumn)
  {
    return this.m_columns.indexOf(paramColumn);
  }
  
  public Column getColumn(int paramInt)
  {
    this.m_lastCol = paramInt;
    return (Column)this.m_columns.get(paramInt);
  }
  
  public Column getColumn(String paramString)
  {
    ColumnEntry localColumnEntry = (ColumnEntry)this.m_entries.get(paramString);
    return localColumnEntry != null ? localColumnEntry.column : null;
  }
  
  public void addColumn(String paramString, Class paramClass)
  {
    addColumn(paramString, paramClass, null);
  }
  
  public void addColumn(String paramString, Class paramClass, Object paramObject)
  {
    Column localColumn = ColumnFactory.getColumn(paramClass, this.m_rows.getMaximumRow() + 1, paramObject);
    addColumn(paramString, localColumn);
  }
  
  public void addColumn(String paramString1, String paramString2)
  {
    Expression localExpression = ExpressionParser.parse(paramString2);
    Throwable localThrowable = ExpressionParser.getError();
    if (localThrowable != null) {
      throw new RuntimeException(localThrowable);
    }
    addColumn(paramString1, localExpression);
  }
  
  public void addColumn(String paramString, Expression paramExpression)
  {
    addColumn(paramString, ColumnFactory.getColumn(this, paramExpression));
  }
  
  public void addConstantColumn(String paramString, Class paramClass, Object paramObject)
  {
    addColumn(paramString, ColumnFactory.getConstantColumn(paramClass, paramObject));
  }
  
  protected void addColumn(String paramString, Column paramColumn)
  {
    int i = getColumnNumber(paramString);
    if ((i >= 0) && (i < this.m_columns.size())) {
      throw new IllegalArgumentException("Table already has column with name \"" + paramString + "\"");
    }
    this.m_columns.add(paramColumn);
    this.m_names.add(paramString);
    this.m_lastCol = (this.m_columns.size() - 1);
    ColumnEntry localColumnEntry1 = new ColumnEntry(this.m_lastCol, paramColumn, new ColumnMetadata(this, paramString));
    ColumnEntry localColumnEntry2 = (ColumnEntry)this.m_entries.put(paramString, localColumnEntry1);
    if (localColumnEntry2 != null) {
      localColumnEntry2.dispose();
    }
    invalidateSchema();
    paramColumn.addColumnListener(this);
    fireTableEvent(this.m_rows.getMinimumRow(), this.m_rows.getMaximumRow(), this.m_lastCol, 1);
  }
  
  protected Column removeColumn(int paramInt)
  {
    if ((paramInt < 0) || (paramInt >= this.m_columns.size())) {
      throw new IllegalArgumentException("Column index is not legal.");
    }
    String str = (String)this.m_names.get(paramInt);
    ((ColumnEntry)this.m_entries.get(str)).dispose();
    Column localColumn = (Column)this.m_columns.remove(paramInt);
    this.m_entries.remove(str);
    this.m_names.remove(paramInt);
    renumberColumns();
    this.m_lastCol = -1;
    invalidateSchema();
    localColumn.removeColumnListener(this);
    fireTableEvent(this.m_rows.getMinimumRow(), this.m_rows.getMaximumRow(), paramInt, -1);
    return localColumn;
  }
  
  public Column removeColumn(String paramString)
  {
    int i = this.m_names.indexOf(paramString);
    if (i < 0) {
      throw new IllegalArgumentException("No such column.");
    }
    return removeColumn(i);
  }
  
  public void removeColumn(Column paramColumn)
  {
    int i = this.m_columns.indexOf(paramColumn);
    if (i < 0) {
      throw new IllegalArgumentException("No such column.");
    }
    removeColumn(i);
  }
  
  protected void renumberColumns()
  {
    Iterator localIterator = this.m_names.iterator();
    for (int i = 0; localIterator.hasNext(); i++)
    {
      String str = (String)localIterator.next();
      ColumnEntry localColumnEntry = (ColumnEntry)this.m_entries.get(str);
      localColumnEntry.colnum = i;
    }
  }
  
  protected Iterator getColumns()
  {
    return this.m_columns.iterator();
  }
  
  protected Iterator getColumnNames()
  {
    return this.m_names.iterator();
  }
  
  public ColumnMetadata getMetadata(String paramString)
  {
    ColumnEntry localColumnEntry = (ColumnEntry)this.m_entries.get(paramString);
    if (localColumnEntry == null) {
      throw new IllegalArgumentException("Unknown column name: " + paramString);
    }
    return localColumnEntry.metadata;
  }
  
  public Index index(String paramString)
  {
    ColumnEntry localColumnEntry = (ColumnEntry)this.m_entries.get(paramString);
    if (localColumnEntry == null) {
      throw new IllegalArgumentException("Unknown column name: " + paramString);
    }
    if (localColumnEntry.index != null) {
      return localColumnEntry.index;
    }
    Column localColumn = localColumnEntry.column;
    try
    {
      localColumnEntry.index = new TreeIndex(this, this.m_rows, localColumn, null);
    }
    catch (IncompatibleComparatorException localIncompatibleComparatorException) {}
    return localColumnEntry.index;
  }
  
  public Index getIndex(String paramString)
  {
    ColumnEntry localColumnEntry = (ColumnEntry)this.m_entries.get(paramString);
    if (localColumnEntry == null) {
      throw new IllegalArgumentException("Unknown column name: " + paramString);
    }
    return localColumnEntry.index;
  }
  
  protected Index getIndex(String paramString, Class paramClass, boolean paramBoolean)
  {
    if (!paramClass.equals(getColumnType(paramString))) {
      throw new IllegalArgumentException("Column type does not match.");
    }
    if ((getIndex(paramString) == null) && (paramBoolean)) {
      index(paramString);
    }
    return getIndex(paramString);
  }
  
  public boolean removeIndex(String paramString)
  {
    ColumnEntry localColumnEntry = (ColumnEntry)this.m_entries.get(paramString);
    if (localColumnEntry == null) {
      throw new IllegalArgumentException("Unknown column name: " + paramString);
    }
    if (localColumnEntry.index == null) {
      return false;
    }
    localColumnEntry.index.dispose();
    localColumnEntry.index = null;
    return true;
  }
  
  public Tuple getTuple(int paramInt)
  {
    return this.m_tuples.getTuple(paramInt);
  }
  
  public Tuple addTuple(Tuple paramTuple)
  {
    if (paramTuple.getTable() == this) {
      return null;
    }
    Schema localSchema = paramTuple.getSchema();
    if (getSchema().isAssignableFrom(localSchema))
    {
      int i = addRow();
      for (int j = 0; j < localSchema.getColumnCount(); j++)
      {
        String str = localSchema.getColumnName(j);
        set(i, str, paramTuple.getValueAt(j));
      }
      return getTuple(i);
    }
    return null;
  }
  
  public Tuple setTuple(Tuple paramTuple)
  {
    clear();
    return addTuple(paramTuple);
  }
  
  public boolean removeTuple(Tuple paramTuple)
  {
    if (containsTuple(paramTuple))
    {
      removeRow(paramTuple.getRow());
      return true;
    }
    return false;
  }
  
  public boolean containsTuple(Tuple paramTuple)
  {
    return (paramTuple.getTable() == this) && (isValidRow(paramTuple.getRow()));
  }
  
  public int getTupleCount()
  {
    return getRowCount();
  }
  
  public boolean isAddColumnSupported()
  {
    return true;
  }
  
  public boolean canGet(String paramString, Class paramClass)
  {
    Column localColumn = getColumn(paramString);
    return localColumn == null ? false : localColumn.canGet(paramClass);
  }
  
  public boolean canSet(String paramString, Class paramClass)
  {
    Column localColumn = getColumn(paramString);
    return localColumn == null ? false : localColumn.canSet(paramClass);
  }
  
  public Object get(int paramInt, String paramString)
  {
    int i = getColumnNumber(paramString);
    paramInt = getColumnRow(paramInt, i);
    return getColumn(i).get(paramInt);
  }
  
  public void set(int paramInt, String paramString, Object paramObject)
  {
    int i = getColumnNumber(paramString);
    paramInt = getColumnRow(paramInt, i);
    getColumn(i).set(paramObject, paramInt);
  }
  
  public Object getValueAt(int paramInt1, int paramInt2)
  {
    paramInt1 = getColumnRow(paramInt1, paramInt2);
    return getColumn(paramInt2).get(paramInt1);
  }
  
  public void setValueAt(int paramInt1, int paramInt2, Object paramObject)
  {
    paramInt1 = getColumnRow(paramInt1, paramInt2);
    getColumn(paramInt2).set(paramObject, paramInt1);
  }
  
  public Object getDefault(String paramString)
  {
    int i = getColumnNumber(paramString);
    return getColumn(i).getDefaultValue();
  }
  
  public void revertToDefault(int paramInt, String paramString)
  {
    int i = getColumnNumber(paramString);
    paramInt = getColumnRow(paramInt, i);
    getColumn(i).revertToDefault(paramInt);
  }
  
  public final boolean canGetInt(String paramString)
  {
    Column localColumn = getColumn(paramString);
    return localColumn == null ? false : localColumn.canGetInt();
  }
  
  public final boolean canSetInt(String paramString)
  {
    Column localColumn = getColumn(paramString);
    return localColumn == null ? false : localColumn.canSetInt();
  }
  
  public final int getInt(int paramInt, String paramString)
  {
    int i = getColumnNumber(paramString);
    paramInt = getColumnRow(paramInt, i);
    return getColumn(i).getInt(paramInt);
  }
  
  public final void setInt(int paramInt1, String paramString, int paramInt2)
  {
    int i = getColumnNumber(paramString);
    paramInt1 = getColumnRow(paramInt1, i);
    getColumn(i).setInt(paramInt2, paramInt1);
  }
  
  public final boolean canGetLong(String paramString)
  {
    Column localColumn = getColumn(paramString);
    return localColumn == null ? false : localColumn.canGetLong();
  }
  
  public final boolean canSetLong(String paramString)
  {
    Column localColumn = getColumn(paramString);
    return localColumn == null ? false : localColumn.canSetLong();
  }
  
  public final long getLong(int paramInt, String paramString)
  {
    int i = getColumnNumber(paramString);
    paramInt = getColumnRow(paramInt, i);
    return getColumn(i).getLong(paramInt);
  }
  
  public final void setLong(int paramInt, String paramString, long paramLong)
  {
    int i = getColumnNumber(paramString);
    paramInt = getColumnRow(paramInt, i);
    getColumn(i).setLong(paramLong, paramInt);
  }
  
  public final boolean canGetFloat(String paramString)
  {
    Column localColumn = getColumn(paramString);
    return localColumn == null ? false : localColumn.canGetFloat();
  }
  
  public final boolean canSetFloat(String paramString)
  {
    Column localColumn = getColumn(paramString);
    return localColumn == null ? false : localColumn.canSetFloat();
  }
  
  public final float getFloat(int paramInt, String paramString)
  {
    int i = getColumnNumber(paramString);
    paramInt = getColumnRow(paramInt, i);
    return getColumn(i).getFloat(paramInt);
  }
  
  public final void setFloat(int paramInt, String paramString, float paramFloat)
  {
    int i = getColumnNumber(paramString);
    paramInt = getColumnRow(paramInt, i);
    getColumn(i).setFloat(paramFloat, paramInt);
  }
  
  public final boolean canGetDouble(String paramString)
  {
    Column localColumn = getColumn(paramString);
    return localColumn == null ? false : localColumn.canGetDouble();
  }
  
  public final boolean canSetDouble(String paramString)
  {
    Column localColumn = getColumn(paramString);
    return localColumn == null ? false : localColumn.canSetDouble();
  }
  
  public final double getDouble(int paramInt, String paramString)
  {
    int i = getColumnNumber(paramString);
    paramInt = getColumnRow(paramInt, i);
    return getColumn(i).getDouble(paramInt);
  }
  
  public final void setDouble(int paramInt, String paramString, double paramDouble)
  {
    int i = getColumnNumber(paramString);
    paramInt = getColumnRow(paramInt, i);
    getColumn(i).setDouble(paramDouble, paramInt);
  }
  
  public final boolean canGetBoolean(String paramString)
  {
    Column localColumn = getColumn(paramString);
    return localColumn == null ? false : localColumn.canGetBoolean();
  }
  
  public final boolean canSetBoolean(String paramString)
  {
    Column localColumn = getColumn(paramString);
    return localColumn == null ? false : localColumn.canSetBoolean();
  }
  
  public final boolean getBoolean(int paramInt, String paramString)
  {
    int i = getColumnNumber(paramString);
    paramInt = getColumnRow(paramInt, i);
    return getColumn(i).getBoolean(paramInt);
  }
  
  public final void setBoolean(int paramInt, String paramString, boolean paramBoolean)
  {
    int i = getColumnNumber(paramString);
    paramInt = getColumnRow(paramInt, i);
    getColumn(i).setBoolean(paramBoolean, paramInt);
  }
  
  public final boolean canGetString(String paramString)
  {
    Column localColumn = getColumn(paramString);
    return localColumn == null ? false : localColumn.canGetString();
  }
  
  public final boolean canSetString(String paramString)
  {
    Column localColumn = getColumn(paramString);
    return localColumn == null ? false : localColumn.canSetString();
  }
  
  public final String getString(int paramInt, String paramString)
  {
    int i = getColumnNumber(paramString);
    paramInt = getColumnRow(paramInt, i);
    return getColumn(i).getString(paramInt);
  }
  
  public final void setString(int paramInt, String paramString1, String paramString2)
  {
    int i = getColumnNumber(paramString1);
    paramInt = getColumnRow(paramInt, i);
    getColumn(i).setString(paramString2, paramInt);
  }
  
  public final boolean canGetDate(String paramString)
  {
    Column localColumn = getColumn(paramString);
    return localColumn == null ? false : localColumn.canGetDate();
  }
  
  public final boolean canSetDate(String paramString)
  {
    Column localColumn = getColumn(paramString);
    return localColumn == null ? false : localColumn.canSetDate();
  }
  
  public final Date getDate(int paramInt, String paramString)
  {
    int i = getColumnNumber(paramString);
    paramInt = getColumnRow(paramInt, i);
    return getColumn(i).getDate(paramInt);
  }
  
  public final void setDate(int paramInt, String paramString, Date paramDate)
  {
    int i = getColumnNumber(paramString);
    paramInt = getColumnRow(paramInt, i);
    getColumn(i).setDate(paramDate, paramInt);
  }
  
  public TableIterator iterator()
  {
    return iterator(rows());
  }
  
  public TableIterator iterator(IntIterator paramIntIterator)
  {
    return new TableIterator(this, paramIntIterator);
  }
  
  public Iterator tuples()
  {
    return this.m_tuples.iterator(rows());
  }
  
  public Iterator tuplesReversed()
  {
    return this.m_tuples.iterator(rows(true));
  }
  
  public Iterator tuples(IntIterator paramIntIterator)
  {
    return this.m_tuples.iterator(paramIntIterator);
  }
  
  public IntIterator rows()
  {
    return this.m_rows.rows();
  }
  
  public IntIterator rows(Predicate paramPredicate)
  {
    return FilterIteratorFactory.rows(this, paramPredicate);
  }
  
  public IntIterator rows(boolean paramBoolean)
  {
    return this.m_rows.rows(paramBoolean);
  }
  
  public IntIterator rowsSortedBy(String paramString, boolean paramBoolean)
  {
    Class localClass = getColumnType(paramString);
    Index localIndex = getIndex(paramString, localClass, true);
    int i = paramBoolean ? 32 : 16;
    return localIndex.allRows(i);
  }
  
  public IntIterator rangeSortedBy(String paramString, int paramInt1, int paramInt2, int paramInt3)
  {
    Index localIndex = getIndex(paramString, Integer.TYPE, true);
    return localIndex.rows(paramInt1, paramInt2, paramInt3);
  }
  
  public IntIterator rangeSortedBy(String paramString, long paramLong1, long paramLong2, int paramInt)
  {
    Index localIndex = getIndex(paramString, Long.TYPE, true);
    return localIndex.rows(paramLong1, paramLong2, paramInt);
  }
  
  public IntIterator rangeSortedBy(String paramString, float paramFloat1, float paramFloat2, int paramInt)
  {
    Index localIndex = getIndex(paramString, Float.TYPE, true);
    return localIndex.rows(paramFloat1, paramFloat2, paramInt);
  }
  
  public IntIterator rangeSortedBy(String paramString, double paramDouble1, double paramDouble2, int paramInt)
  {
    Index localIndex = getIndex(paramString, Double.TYPE, true);
    return localIndex.rows(paramDouble1, paramDouble2, paramInt);
  }
  
  public IntIterator rangeSortedBy(String paramString, Object paramObject1, Object paramObject2, int paramInt)
  {
    Class localClass = TypeLib.getSharedType(paramObject1, paramObject2);
    if (localClass == null) {
      throw new IllegalArgumentException("Incompatible arguments");
    }
    Index localIndex = getIndex(paramString, localClass, true);
    return localIndex.rows(paramObject1, paramObject2, paramInt);
  }
  
  public void columnChanged(Column paramColumn, int paramInt, boolean paramBoolean)
  {
    handleColumnChanged(paramColumn, paramInt, paramInt);
  }
  
  public void columnChanged(Column paramColumn, int paramInt, double paramDouble)
  {
    handleColumnChanged(paramColumn, paramInt, paramInt);
  }
  
  public void columnChanged(Column paramColumn, int paramInt, float paramFloat)
  {
    handleColumnChanged(paramColumn, paramInt, paramInt);
  }
  
  public void columnChanged(Column paramColumn, int paramInt1, int paramInt2)
  {
    handleColumnChanged(paramColumn, paramInt1, paramInt1);
  }
  
  public void columnChanged(Column paramColumn, int paramInt, long paramLong)
  {
    handleColumnChanged(paramColumn, paramInt, paramInt);
  }
  
  public void columnChanged(Column paramColumn, int paramInt, Object paramObject)
  {
    handleColumnChanged(paramColumn, paramInt, paramInt);
  }
  
  public void columnChanged(Column paramColumn, int paramInt1, int paramInt2, int paramInt3)
  {
    handleColumnChanged(paramColumn, paramInt2, paramInt3);
  }
  
  protected void handleColumnChanged(Column paramColumn, int paramInt1, int paramInt2)
  {
    while ((!isValidRow(paramInt1)) && (paramInt1 <= paramInt2)) {
      paramInt1++;
    }
    if (paramInt1 > paramInt2) {
      return;
    }
    int i;
    if ((this.m_lastCol != -1) && (paramColumn == getColumn(this.m_lastCol))) {
      i = this.m_lastCol;
    } else {
      i = getColumnNumber(paramColumn);
    }
    if (i >= 0) {
      fireTableEvent(paramInt1, paramInt2, i, 0);
    }
  }
  
  public void addTableListener(TableListener paramTableListener)
  {
    if (!this.m_listeners.contains(paramTableListener)) {
      this.m_listeners.add(paramTableListener);
    }
  }
  
  public void removeTableListener(TableListener paramTableListener)
  {
    this.m_listeners.remove(paramTableListener);
  }
  
  protected void fireTableEvent(int paramInt1, int paramInt2, int paramInt3, int paramInt4)
  {
    this.m_modCount += 1;
    if ((paramInt4 != 0) && (paramInt3 == -1)) {
      fireTupleEvent(this, paramInt1, paramInt2, paramInt4);
    }
    if (!this.m_listeners.isEmpty())
    {
      Object[] arrayOfObject = this.m_listeners.getArray();
      for (int i = 0; i < arrayOfObject.length; i++) {
        ((TableListener)arrayOfObject[i]).tableChanged(this, paramInt1, paramInt2, paramInt3, paramInt4);
      }
    }
  }
  
  public String toString()
  {
    StringBuffer localStringBuffer = new StringBuffer();
    localStringBuffer.append("Table[");
    localStringBuffer.append("rows=").append(getRowCount());
    localStringBuffer.append(", cols=").append(getColumnCount());
    localStringBuffer.append(", maxrow=").append(this.m_rows.getMaximumRow());
    localStringBuffer.append("]");
    return localStringBuffer.toString();
  }
  
  protected static class ColumnEntry
  {
    public int colnum;
    public Column column;
    public ColumnMetadata metadata;
    public Index index;
    
    public ColumnEntry(int paramInt, Column paramColumn, ColumnMetadata paramColumnMetadata)
    {
      this.colnum = paramInt;
      this.column = paramColumn;
      this.metadata = paramColumnMetadata;
      this.index = null;
    }
    
    public void dispose()
    {
      if (this.metadata != null) {
        this.metadata.dispose();
      }
      if (this.index != null) {
        this.index.dispose();
      }
    }
  }
}


/* Location:              C:\Perforce\MyAsup\dev\thirdparty\lib\asup-metamodel-client-1.0.jar!\prefuse\data\Table.class
 * Java compiler version: 5 (49.0)
 * JD-Core Version:       0.7.1
 */